(**************************************************************************)
(*                     NFA' are special cases of NFE'                     *)
(**************************************************************************)

Implicit Arguments On.

Require MoreProperties.
Require NFA.
Require NFA_eps.

Section nfa'_and_nfe'.

Variable A,St : Set.
Variable I    : (fset A).
Variable eps  : A.
Hypothesis eps_not_in_I : ~(member eps I).

(* If there is a path in a NFA A then there is the same path in A
   seen as a NFE. *)

Lemma nfa'_path_is_nfe'_path :
  (Q:(fset St))(delta:(fset St*A*St))(q0:St)
  (f:St)(w:(word A))
  (nf_path I Q delta q0 f w) -> (nfe_path eps I Q delta q0 f w).

(* Conversly, if there is a path in a NFE A, and if there is no epsilon
   transition in delta, then there is the same path in A seen as a NFA. *)

Lemma nfe'_path_is_nfa'_path :
  (Q:(fset St))(delta:(fset St*A*St))(q0:St)
  (f:St)(w:(word A))
  (subset delta (product Q (product I Q))) ->
  (nfe_path eps I Q delta q0 f w) -> (nf_path I Q delta q0 f w).

(* Using the two previous results, we can prove that a language L recognized
   by a NFA' is recognized by a NFE'. It happens to be the same automaton. *)

Lemma nfa'_regular_is_nfe'_regular :
  (L:(wordset A)) (nf'_regularset St I L) -> (nfe'_regularset St eps I L).

End nfa'_and_nfe'.

