(**************************************************************************)
(*                     NFA' are special cases of NFE'                     *)
(**************************************************************************)

Implicit Arguments On.

Require MoreProperties.
Require NFA.
Require NFA_eps.


Section nfa'_and_nfe'.

Variable A,St : Set.
Variable I    : (fset A).
Variable eps  : A.
Hypothesis eps_not_in_I : ~(member eps I).


(* If there is a path in a NFA A then there is the same path in A
   seen as a NFE. *)

Lemma nfa'_path_is_nfe'_path :
  (Q:(fset St))(delta:(fset St*A*St))(q0:St)
  (f:St)(w:(word A))
  (nf_path I Q delta q0 f w) -> (nfe_path eps I Q delta q0 f w).
Proof.
Intros Q delta q0 f w.
Induction 1.

Intros e H_e.
Apply nfe_path_nil.
Assumption.

Intros e1 e2 e w' a H_w H_nfe_path_w H_a H_e_a_e1 H_e.
Apply nfe_path_cons with e1:=e1 ; Auto.
Save.


(* Conversly, if there is a path in a NFE A, and if there is no epsilon
   transition in delta, then there is the same path in A seen as a NFA. *)

Lemma nfe'_path_is_nfa'_path :
  (Q:(fset St))(delta:(fset St*A*St))(q0:St)
  (f:St)(w:(word A))
  (subset delta (product Q (product I Q))) ->
  (nfe_path eps I Q delta q0 f w) -> (nf_path I Q delta q0 f w).
Proof.
Intros Q delta q0 f w H_delta.
Induction 1.

Intros e H_e.
Apply nf_path_nil.
Assumption.

Intros e1 e2 e w' a H_w H_nf_path_w H_a H_e_a_e1 H_e.
Apply nf_path_cons with e1:=e1 ; Auto.

Intros e1 e2 e w' H_w H_nf_path_w H_eps H_e.
Absurd (member eps I) ; Auto.
Apply prod_left with b:=Q xb:=e1.
Apply prod_right with a:=Q xa:=e.
Apply member_subset with a:=delta ; Assumption.
Save.


(* Using the two previous results, we can prove that a language L recognized
   by a NFA' is recognized by a NFE'. It happens to be the same automaton. *)

Lemma nfa'_regular_is_nfe'_regular :
  (L:(wordset A)) (nf'_regularset St I L) -> (nfe'_regularset St eps I L).
Proof.
Intros L H_reg.
Elim H_reg.
Intros Q q0 f delta H_aut H_eq.
Apply nfe'_regular_intro with Q:=Q q0:=q0 f:=f delta:=delta.

Elim H_aut.
Intros H_f H_q0 H_f_delta H_delta.
Apply nfe'_automaton_intro ; Auto.
Apply subset_prod32. Assumption.

Unfold eq_wordset in H_eq. Unfold eq_wordset.
Intro w. Elim (H_eq w).
Intros H_w_1 H_w_2.
Split.

Intro L_w. Elim (H_w_1 L_w).
Intros f' H_f' H_path.
Apply nfe_accepts_intro with f:=f'.
Assumption.
Apply nfa'_path_is_nfe'_path. Assumption.

Intro H_acc.
Apply H_w_2.
Elim H_acc.
Intros f' H_f' H_path.
Apply nf_accepts_intro with f:=f'.
Assumption.
Apply nfe'_path_is_nfa'_path.
Elim H_aut. Auto.
Assumption.
Save.

End nfa'_and_nfe'.

