
(**************************************************************************)
(* Equivalence of automata with states over naturals and with             *)
(* states over naturals x booleans.                                       *)
(**************************************************************************)

Implicit Arguments On.

Require Export DFA.
Require Export NFA.
Require Export NFA_eps.
Require NatProperties.

Section Equiv.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).

Section Aut_for_equiv.

Variable l1 : (wordset A).

Variable Q1 : (fset nat*bool).
Variable q0_1 : nat*bool.
Variable f1 : nat*bool.
Variable delta1 : (fset (nat*bool)*A*(nat*bool)).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).

Definition m := (proj1_sig ? ? (outside (map [x:nat*bool](Fst x) Q1))).

Definition f : nat*bool -> nat :=
  [x:nat*bool]
  Cases x of (pair n b) =>
    Cases b of false => n
             | true  => (plus m n)
    end
  end.

Definition Q := (map f Q1).

Definition q0 := (f q0_1).

Definition f0 : nat := (f f1).

Definition delta :=
  (map [t:(nat*bool)*A*(nat*bool)]
       Cases t of (pair x c) =>
         Cases c of (pair q x') => (f x,(q,f x'))
       end end
  delta1).

Lemma member_f0_Q : (member f0 Q).

Lemma member_q0_Q : (member q0 Q).

(* f is a one to one mapping on Q1. *)

Axiom injection : (q,q':nat*bool)
  (member q Q1) ->
  (member q' Q1) ->
  (f q)=(f q') -> q=q'.

Lemma H_f0 : (a:A)(q:nat)~(member (f0,(a,q)) delta).

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).

Lemma is_automaton : (nfe'_automaton eps I Q q0 f0 delta).

Axiom eq_accepts :
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l1 (nfe_accepts eps I Q  q0   (singleton f0) delta)).

End Aut_for_equiv.

Lemma natbool_equiv_nat : 
  (l1:(wordset A))
  (nfe'_regularset nat*bool eps I l1) -> (nfe'_regularset nat eps I l1).

End Equiv.
