
(**************************************************************************)
(* Equivalence of automata with states over naturals and with             *)
(* states over naturals x booleans.                                       *)
(**************************************************************************)

Implicit Arguments On.

Require Export DFA.
Require Export NFA.
Require Export NFA_eps.
Require NatProperties.


Section Equiv.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).


Section Aut_for_equiv.

Variable l1 : (wordset A).

Variable Q1 : (fset nat*bool).
Variable q0_1 : nat*bool.
Variable f1 : nat*bool.
Variable delta1 : (fset (nat*bool)*A*(nat*bool)).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).


Definition m := (proj1_sig ? ? (outside (map [x:nat*bool](Fst x) Q1))).

Definition f : nat*bool -> nat := 
  [x:nat*bool]
  Cases x of (pair n b) => 
    Cases b of false => n
             | true  => (plus m n)
    end
  end.

Definition Q := (map f Q1).

Definition q0 := (f q0_1).

Definition f0 : nat := (f f1).

Definition delta :=
  (map [t:(nat*bool)*A*(nat*bool)]
       Cases t of (pair x c) =>
         Cases c of (pair q x') => (f x,(q,f x'))
       end end
  delta1).


Lemma member_f0_Q : (member f0 Q).
Proof.
Unfold f0 Q.
Apply map_def_1.
(Elim aut1; Auto).
Save.

Lemma member_q0_Q : (member q0 Q).
Proof.
Unfold q0 Q.
Apply map_def_1.
(Elim aut1; Auto).
Save.

(* f is a one to one mapping on Q1. *)

Axiom injection : (q,q':nat*bool)
  (member q Q1) ->
  (member q' Q1) ->
  (f q)=(f q') -> q=q'.

Lemma H_f0 : (a:A)(q:nat)~(member (f0,(a,q)) delta).
Proof.
Unfold delta.
Red.
Intros a q Hm.
Elim (map_def_2 Hm).
Intros x.
Case x.
Intros p x'.
Case x'.
Intros a0 p' H.
Elim H.
Intros Hm' equ.
Injection equ.
Intros equ1 equ2 equ3.
Elim aut1.
Intros.
Apply (H3 a0 p').
Unfold f0 in equ3.
Cut f1=p.
Intros equ'.
Rewrite -> equ'.
Assumption.
 
Apply injection.
Assumption.
 
Elim H.
Intros.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a0,p').
Apply member_subset with a:=delta1.
Assumption.
 
Assumption.
 
Assumption.
Save.

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).
Proof.
Unfold subset delta.
Intros x Hm.
Elim (map_def_2 Hm).
Clear  Hm.
Induction x0.
Intros q.
Induction y.
Intros a q'.
Intros H.
Elim H.
Clear  H.
Intros Hm equ.
Rewrite -> equ.
Unfold Q.
Apply member_pair_prod.
Apply map_def_1.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a,q').
Apply member_subset with a:=delta1.
Assumption.
 
Elim aut1.
Auto.
 
Apply member_pair_prod.
Apply prod_left with b:=Q1 xb:=q'.
Apply prod_right with a:=Q1 xa:=q.
Apply member_subset with a:=delta1.
Assumption.
 
Elim aut1.
Auto.
 
Apply map_def_1.
Apply prod_right with a:=(add eps I) xa:=a.
Apply prod_right with a:=Q1 xa:=q.
Apply member_subset with a:=delta1.
Assumption.
 
Elim aut1.
Auto.
Save.


Lemma is_automaton : (nfe'_automaton eps I Q q0 f0 delta).
Proof.
Apply (nfe'_automaton_intro 
      	 member_f0_Q
      	 eps_not_in_I 
      	 member_q0_Q
      	 H_f0
      	 H_delta).
Save.


Axiom eq_accepts : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l1 (nfe_accepts eps I Q  q0   (singleton f0) delta)).

End Aut_for_equiv.


Lemma natbool_equiv_nat : 
  (l1:(wordset A))
  (nfe'_regularset nat*bool eps I l1) -> (nfe'_regularset nat eps I l1).
Proof.
Intros l1 nfe'_l1.
Elim nfe'_l1. Intros Q1 q0_1 f1 delta1 aut1 H_l1.
Apply nfe'_regular_intro
  with Q  := (Q Q1)
       q0 := (q0 Q1 q0_1)
       f  := (f0 Q1 f1)
       delta := (delta Q1 delta1).
Apply is_automaton ; Assumption.
Apply eq_accepts ; Assumption.
Save.

End Equiv.
