
(**************************************************************************)
(*                   Preamble for the extracted program                   *)
(*                             (done by hand)                             *)
(**************************************************************************)

type input = char option;;

let symbol_a = Some 'a'
and symbol_b = Some 'b'
and symbol_c = Some 'c';;

let (epsilon : char option) = None;;


(**************************************************************************)
(*                     The extracted program "as is"                      *)
(**************************************************************************)

type bool =
    True
  | False
;;

type nat =
    O
  | S of nat
;;

type ('a, 'b) prod =
    Pair of 'a * 'b
;;

let rec plus n m =
  match n with
    O -> m
  | S p -> S (plus p m)
;;

type 'a list =
    Nil
  | Cons of 'a * ('a list)
;;

let rec app l m =
  match l with
    Nil -> m
  | Cons (a0, l1) -> Cons (a0, (app l1 m))
;;

type 'a fset = 'a list;;

let rec map f = function
  Nil -> Nil
| Cons (a0, la') -> Cons ((f a0), (map f la'))
;;

type 'a word = 'a list;;

type 't regexp =
    Reg_empty
  | Reg_word of ('t word)
  | Reg_union of ('t regexp) * ('t regexp)
  | Reg_conc of ('t regexp) * ('t regexp)
  | Reg_star of ('t regexp)
;;

type ('a, 'st) nfe'_regularset =
    Nfe'_regular_intro of ('st fset) * 'st * 'st
      * (('st, ('a, 'st) prod) prod fset)
;;

let rec max n m =
  match n with
    O -> m
  | S n' -> (match m with
               O -> S n'
             | S m' -> S (max n' m'))
;;

let rec max_of_list = function
  Nil -> O
| Cons (x, l') -> max x (max_of_list l')
;;

let outside2 e =
  Pair ((S (max_of_list
             (map (fun x -> match x with
                              Pair (x0, y) -> x0) e))), True)
;;

type ('a, 'st) nf'_regularset =
    Nf'_regular_intro of ('st fset) * 'st * 'st
      * (('st, ('a, 'st) prod) prod fset)
;;

let disj_union a0 b =
  app (map (fun x -> Pair (x, False)) a0)
    (map (fun x -> Pair (x, True)) b)
;;

let exists_q0 q1 q2 =
  outside2 (disj_union q1 q2)
;;

let exists_f0 q1 q2 =
  outside2 (Cons ((exists_q0 q1 q2), (disj_union q1 q2)))
;;

let q q1 q2 =
  Cons ((exists_q0 q1 q2),
        (Cons ((exists_f0 q1 q2), (disj_union q1 q2))))
;;

let delta'1 delta1 =
  map (fun x ->
    match x with
      Pair (q1, a0) ->
        (match a0 with
           Pair (a1, q') ->
             Pair ((Pair (q1, False)), (Pair (a1, (Pair (q', False)))))))
    delta1
;;

let delta'2 delta2 =
  map (fun x ->
    match x with
      Pair (q1, a0) ->
        (match a0 with
           Pair (a1, q') ->
             Pair ((Pair (q1, True)), (Pair (a1, (Pair (q', True)))))))
    delta2
;;

let delta eps q1 q2 q0_1 q0_2 f1 f2 delta1 delta2 =
  Cons ((Pair ((exists_q0 q1 q2), (Pair (eps, (Pair (q0_1, False)))))),
        (Cons ((Pair ((exists_q0 q1 q2),
                      (Pair (eps, (Pair (q0_2, True)))))),
               (Cons ((Pair ((Pair (f1, False)),
                             (Pair (eps, (exists_f0 q1 q2))))),
                      (Cons ((Pair ((Pair (f2, True)),
                                    (Pair (eps, (exists_f0 q1 q2))))),
                             (app (delta'1 delta1) (delta'2 delta2)))))))))
;;

let closed_under_union i eps reg_l1 nfe'_l1 reg_l2 nfe'_l2 =
  match nfe'_l1 with
    Nfe'_regular_intro (x2, x1, x0, x) ->
      (match nfe'_l2 with
         Nfe'_regular_intro (x6, x5, x4, x3) ->
           Nfe'_regular_intro ((q x2 x6), (exists_q0 x2 x6),
                               (exists_f0 x2 x6),
                               (delta eps x2 x6 x1 x5 x0 x4 x x3)))
;;

let delta'0 delta1 =
  map (fun x ->
    match x with
      Pair (q3, a0) ->
        (match a0 with
           Pair (a1, q') ->
             Pair ((Pair (q3, False)), (Pair (a1, (Pair (q', False)))))))
    delta1
;;

let delta'3 delta2 =
  map (fun x ->
    match x with
      Pair (q3, a0) ->
        (match a0 with
           Pair (a1, q') ->
             Pair ((Pair (q3, True)), (Pair (a1, (Pair (q', True)))))))
    delta2
;;

let delta0 eps q0_2 f2 delta1 delta2 =
  Cons ((Pair ((Pair (f2, False)), (Pair (eps, (Pair (q0_2, True)))))),
        (app (delta'0 delta1) (delta'3 delta2)))
;;

let delta1 eps q5 q0_1 f3 delta1 =
  Cons ((Pair (q0_1, (Pair (eps, (S (max_of_list q5)))))),
        (Cons ((Pair (f3, (Pair (eps, q0_1)))),
               (Cons ((Pair (f3, (Pair (eps, (S (max_of_list q5)))))),
                      delta1)))))
;;

let m q5 =
  S (max_of_list (map (fun x -> match x with
                                  Pair (x0, y) -> x0) q5))
;;

let q5 q5 =
  map (fun x ->
    match x with
      Pair (n, b) -> (match b with
                        True -> plus (m q5) n
                      | False -> n)) q5
;;

let natbool_equiv_nat i eps = function
  Nfe'_regular_intro (x2, x1, x0, x) ->
    Nfe'_regular_intro ((q5 x2),
                        (match x1 with
                           Pair (n0, b) ->
                             (match b with
                                True -> plus (m x2) n0
                              | False -> n0)),
                        (match x0 with
                           Pair (n0, b) ->
                             (match b with
                                True -> plus (m x2) n0
                              | False -> n0)),
                        (map (fun t ->
                          match t with
                            Pair (x3, c) ->
                              (match c with
                                 Pair (q8, x') ->
                                   Pair ((match x3 with
                                            Pair (n0, b) ->
                                              (match b with
                                                 True -> plus (m x2) n0
                                               | False -> n0)),
                                         (Pair (q8,
                                                (match x' with
                                                   Pair (n0, b) ->
                                                     (match b with
                                                        True ->
                                                          plus 
                                                            (m x2) n0
                                                      | False -> n0)))))))
                          x))
;;

type a = input;;

let regexp_nfe'_automaton l_reg =
  let rec f4 = function
    Reg_empty ->
      Nfe'_regular_intro ((Cons ((S O), (Cons (O, Nil)))), O, (S O),
                          Nil)
  | Reg_word w ->
      (match let rec f0 = function
               Nil -> Nf'_regular_intro ((Cons (O, Nil)), O, O, Nil)
             | Cons (a0, l0) ->
                 (match f0 l0 with
                    Nf'_regular_intro (x2, x1, x0, x) ->
                      Nf'_regular_intro ((Cons ((S (max_of_list x2)),
                                                x2)),
                                         (S (max_of_list x2)), x0,
                                         (Cons ((Pair ((S (max_of_list
                                                            x2)),
                                                       (Pair (a0, x1)))),
                                                x))))
             in f0 w with
         Nf'_regular_intro (x2, x1, x0, x) ->
           Nfe'_regular_intro (x2, x1, x0, x))
  | Reg_union (r0, r1) ->
      natbool_equiv_nat (Cons (symbol_a, (Cons (symbol_b, Nil)))) None
        (closed_under_union (Cons (symbol_a, (Cons (symbol_b, Nil))))
          None r0 (f4 r0) r1 (f4 r1))
  | Reg_conc (r0, r1) ->
      natbool_equiv_nat (Cons (symbol_a, (Cons (symbol_b, Nil)))) None
        (match f4 r0 with
           Nfe'_regular_intro (x2, x1, x0, x) ->
             (match f4 r1 with
                Nfe'_regular_intro (x6, x5, x4, x3) ->
                  Nfe'_regular_intro ((disj_union x2 x6),
                                      (Pair (x1, False)),
                                      (Pair (x4, True)),
                                      (delta0 None x5 x0 x x3))))
  | Reg_star r0 ->
      (match f4 r0 with
         Nfe'_regular_intro (x2, x1, x0, x) ->
           Nfe'_regular_intro ((Cons ((S (max_of_list x2)), x2)), x1,
                               (S (max_of_list x2)),
                               (delta1 None x2 x1 x0 x)))
  in f4 l_reg
;;

