(**************************************************************************)
(*        The family of regular sets if closed under concatenation        *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require DisjUnion.
Require NatProperties.

(* We prove here that the family of regular sets is closed under
   concatenation.							*)

Section Concat.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).

(* Here we define the NFE' automaton that recognizes
   the concatenation of l1 and l2. *)

Section Aut_for_concat.

Variable l1,l2 : (wordset A).

Variable Q1,Q2 : (fset nat).
Variable q0_1,q0_2 : nat.
Variable f1,f2 : nat.
Variable delta1,delta2 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).
Hypothesis aut2 : (nfe'_automaton eps I Q2 q0_2 f2 delta2).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition Q : (fset nat*bool) := (disj_union Q1 Q2).

Definition q0 : nat*bool := (linj q0_1).

Definition f0 : nat*bool := (rinj f2).

Definition delta'1 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (linj q,(a,linj q')) end delta1).

Definition delta'2 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (rinj q,(a,rinj q')) end delta2).

Definition delta : (fset (nat*bool)*A*(nat*bool)) :=
  (add (linj f1,(eps,(rinj q0_2)))
  (union delta'1 delta'2)).

(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).

Lemma member_q0_Q : (member q0 Q).

(**
Lemma f0_not_q0 : ~f0=q0.
Proof.
**)

Require Bool.

Lemma H_f0 : (a:A)(q:nat*bool)~(member (f0,(a,q)) delta).

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).

Lemma concat_automaton : (nfe'_automaton eps I Q q0 f0 delta).

Axiom accepts_concatenation :
  (w1,w2:(word A))
  (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1 w1) ->
  (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2 w2) ->
  (nfe_accepts eps I Q q0 (singleton f0) delta (app A w1 w2)).

Axiom splitting :
  (w:(word A))
  (nfe_path eps I Q delta q0 f0 w) ->
  (Ex [w1:(word A)](Ex [w2:(word A)](
     (nfe_path eps I Q1 delta1 q0_1 f1 w1)
  /\ (nfe_path eps I Q2 delta2 q0_2 f2 w2)
  /\ w=(app ? w1 w2)
  ))).

(* Then we prove that the recognized language is the 
   concatenation of l1 and l2. *)

Lemma concat_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l2 (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2)) ->
  (eq_wordset (wordset_concat l1 l2)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).

End Aut_for_concat.

Theorem closed_under_concat :
  (l1,l2:(wordset A))
  (regexp I l1)
   ->(nfe'_regularset nat eps I l1)
      ->(regexp I l2)
         ->(nfe'_regularset nat eps I l2)
      	   ->(nfe'_regularset nat*bool eps I (wordset_concat l1 l2)).

End Concat.
