(**************************************************************************)
(*        The family of regular sets if closed under concatenation        *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require DisjUnion.
Require NatProperties.


(* We prove here that the family of regular sets is closed under
   concatenation.							*)

Section Concat.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).


(* Here we define the NFE' automaton that recognizes
   the concatenation of l1 and l2. *)

Section Aut_for_concat.

Variable l1,l2 : (wordset A).

Variable Q1,Q2 : (fset nat).
Variable q0_1,q0_2 : nat.
Variable f1,f2 : nat.
Variable delta1,delta2 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).
Hypothesis aut2 : (nfe'_automaton eps I Q2 q0_2 f2 delta2).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition Q : (fset nat*bool) := (disj_union Q1 Q2).

Definition q0 : nat*bool := (linj q0_1).

Definition f0 : nat*bool := (rinj f2).

Definition delta'1 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (linj q,(a,linj q')) end delta1).

Definition delta'2 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (rinj q,(a,rinj q')) end delta2).

Definition delta : (fset (nat*bool)*A*(nat*bool)) :=
  (add (linj f1,(eps,(rinj q0_2)))
  (union delta'1 delta'2)).


(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).
Proof.
Unfold f0 Q.
Apply disj_union_r.
(Elim aut2; Auto).
Save.

Lemma member_q0_Q : (member q0 Q).
Proof.
Unfold q0 Q.
Apply disj_union_l.
(Elim aut1; Auto).
Save.

(**
Lemma f0_not_q0 : ~f0=q0.
Proof.
**)

Require Bool.

Lemma H_f0 : (a:A)(q:nat*bool)~(member (f0,(a,q)) delta).
Proof.
Intros a q.
Red.
Unfold delta.
Intros Hm.
Elim (add_def_1 Hm).

Unfold f0.
Intros equ.
Unfold linj rinj in equ.
Generalize (f_equal ? ? Fst ? ? equ) .
Simpl.
Intros equ'.
Generalize (f_equal ? ? Snd ? ? equ') .
Simpl.
Change ~true=false.
Auto.

Clear  Hm. Intros Hm.
Elim (union_def_1 Hm).
Unfold f0 delta'1.
Clear  Hm. Intros Hm.
Elim (map_def_2 Hm).
Intros x Hc.
Clear  Hm. Elim Hc. Clear  Hc.
Case x.
Intros n p.
Case p.
Intros a0 n0 Hm.
Intros equ.
Generalize (f_equal ? ? Fst ? ? equ) .
Simpl.
Unfold linj rinj.
Intros equ'.
Generalize (f_equal ? ? Snd ? ? equ') .
Simpl.
Auto.

Unfold delta'2.
Clear  Hm.
Intros Hm.
Elim (map_def_2 Hm).
Intros x.
Clear  Hm.
Case x.
Intros n p.
Case p.
Intros a0 n0 Hc.
Elim Hc.
Intros Hm.
Unfold f0.
Intros equ.
Generalize (f_equal ? ? Fst ? ? equ) .
Simpl.
Simpl.
Unfold rinj.
Simpl.
Intros equ'.
Generalize (f_equal ? ? Fst ? ? equ') .
Simpl.
Intros equ''.
Rewrite <- equ'' in Hm.
Elim aut2.
Intros H1 H2 H3 H4 H5.
Apply (H4 a0 n0).
Assumption.
Save.

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).
Proof.
Unfold subset delta.
Intros x Hm. Elim (add_def_1 Hm).

Clear  Hm.
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply disj_union_l.
(Elim aut1; Auto).
Apply member_pair_prod.
Auto.
Unfold Q.
Apply disj_union_r.
(Elim aut2; Auto).

Clear  Hm. Intro Hm. Elim (union_def_1 Hm).

Unfold delta'1.
Clear  Hm. Intros Hm.
Elim (map_def_2 Hm).
Clear  Hm.
Intros x0.
Case x0.
Intros n p.
Case p.
Intros a n0.
Intros Hc. Elim Hc. Clear  Hc. Intros Hm equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply disj_union_l.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a,n0).
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).
Apply member_pair_prod.
Apply prod_left with b:=Q1 xb:=n0.
Apply prod_right with a:=Q1 xa:=n.
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).
Unfold Q.
Apply disj_union_l.
Apply prod_right with a:=(add eps I) xa:=a.
Apply prod_right with a:=Q1 xa:=n.
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).

Unfold delta'2.
Clear  Hm. Intros Hm.
Elim (map_def_2 Hm).
Clear  Hm.
Intros x0.
Case x0.
Intros n p.
Case p.
Intros a n0.
Intros Hc. Elim Hc. Clear  Hc. Intros Hm equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply disj_union_r.
Apply prod_left with b:=(product (add eps I) Q2) xb:=(a,n0).
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Apply member_pair_prod.
Apply prod_left with b:=Q2 xb:=n0.
Apply prod_right with a:=Q2 xa:=n.
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Unfold Q.
Apply disj_union_r.
Apply prod_right with a:=(add eps I) xa:=a.
Apply prod_right with a:=Q2 xa:=n.
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Save.

Lemma concat_automaton : (nfe'_automaton eps I Q q0 f0 delta).
Proof.
Apply (nfe'_automaton_intro 
      	 member_f0_Q
      	 eps_not_in_I 
      	 member_q0_Q
      	 H_f0
      	 H_delta).
Save.


Axiom accepts_concatenation :
  (w1,w2:(word A))
  (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1 w1) ->
  (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2 w2) ->
  (nfe_accepts eps I Q q0 (singleton f0) delta (app A w1 w2)).

Axiom splitting :
  (w:(word A))
  (nfe_path eps I Q delta q0 f0 w) ->
  (Ex [w1:(word A)](Ex [w2:(word A)](
     (nfe_path eps I Q1 delta1 q0_1 f1 w1)
  /\ (nfe_path eps I Q2 delta2 q0_2 f2 w2)
  /\ w=(app ? w1 w2)
  ))).

(* Then we prove that the recognized language is the 
   concatenation of l1 and l2. *) 

Lemma concat_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l2 (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2)) ->
  (eq_wordset (wordset_concat l1 l2)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).
Proof.
Intros H_l1 H_l2.
Unfold eq_wordset.
Intros w.
Split.

(* 1. if a word w is in the concatenation of l1 and l2 then it is accepted. *)
Intros Hu. Elim Hu.
Clear  Hu w. Intros w1 w2 l1_w1 l2_w2.
Unfold eq_wordset in H_l1.
Unfold eq_wordset in H_l2.
Elim (H_l1 w1). Clear H_l1.
Intros H_l1_1 H_l1_2.
Elim (H_l2 w2). Clear H_l2.
Intros H_l2_1 H_l2_2.
Generalize (H_l1_1 l1_w1) .
Generalize (H_l2_1 l2_w2) .
Intros acc2 acc1.
Apply accepts_concatenation ; Assumption.

(* 2. if a word is accepted then it is in the concatenation of l1 and l2. *)
Intros Hacc.
Elim Hacc.
Intros f Hf.
Generalize (member_of_singleton Hf) .
Intros equ.
Rewrite -> equ.
Intro Hpath.
Generalize (splitting Hpath).
Intro ex. Elim ex. Clear ex.
Intros w1 ex. Elim ex. Clear ex.
Intros w2 Hc. Elim Hc. Clear Hc.
Intros Hpath1 Hc. Elim Hc. Clear Hc.
Intros Hpath2 equ'.
Rewrite equ'.
Apply wordset_concat_intro with w1:=w1 w2:=w2.

Unfold eq_wordset in H_l1.
Unfold eq_wordset in H_l2.
Elim (H_l1 w1). Clear H_l1.
Intros H_l1_1 H_l1_2.
Elim (H_l2 w2). Clear H_l2.
Intros H_l2_1 H_l2_2.
Apply H_l1_2.
Apply nfe_accepts_intro with f:=f1.
Auto. Assumption.

Unfold eq_wordset in H_l1.
Unfold eq_wordset in H_l2.
Elim (H_l1 w1). Clear H_l1.
Intros H_l1_1 H_l1_2.
Elim (H_l2 w2). Clear H_l2.
Intros H_l2_1 H_l2_2.
Apply H_l2_2.
Apply nfe_accepts_intro with f:=f2.
Auto. Assumption.
Save.

End Aut_for_concat.


Theorem closed_under_concat :
  (l1,l2:(wordset A))
  (regexp I l1)
   ->(nfe'_regularset nat eps I l1)
      ->(regexp I l2)
         ->(nfe'_regularset nat eps I l2)
      	   ->(nfe'_regularset nat*bool eps I (wordset_concat l1 l2)).
Proof.
Intros l1 l2 reg_l1 nfe'_l1 reg_l2 nfe'_l2.
Elim nfe'_l1. Intros Q1 q0_1 f1 delta1 aut1 H_l1.
Elim nfe'_l2. Intros Q2 q0_2 f2 delta2 aut2 H_l2.
Apply nfe'_regular_intro
  with Q  := (Q Q1 Q2)
       q0 := (q0 q0_1)
       f  := (f0 f2)
       delta := (delta q0_2 f1 delta1 delta2).
Apply concat_automaton ; Assumption.
Apply concat_is_reg ; Assumption.
Save.


End Concat.
