(**************************************************************************)
(*              The family of regular sets if closed under *              *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require NatProperties.
Require MoreProperties.

Section Star.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).

(* Here we define the NFE' automaton that recognizes l* . *)

Section Aut_for_star.

Variable l1 : (wordset A).

Variable Q1 : (fset nat).
Variable q0_1 : nat.
Variable f1 : nat.
Variable delta1 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition exists_f0 := (outside Q1).

Definition f0 : nat := (proj1_sig ? ? exists_f0).

Definition Q := (add f0 Q1).

Definition q0 := q0_1.

Definition delta :=
  (add (q0,(eps,f0))
  (add (f1,(eps,q0))
  (add (f1,(eps,f0))
  delta1))).

(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).

Lemma member_q0_Q : (member q0 Q).

Lemma f0_not_q0 : ~f0=q0.

Lemma f0_not_f1 : ~f0=f1.

Lemma H_f0 : (a:A)(q:nat)~(member (f0,(a,q)) delta).

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).

Lemma star_automaton : (nfe'_automaton eps I Q q0 f0 delta).

Axiom accepts_star :
  (w:(word A))
  (wordset_star l1 w) ->
  (nfe_accepts eps I Q q0 (singleton f0) delta w).

Axiom splitting :
  (w:(word A))
  (nfe_accepts eps I Q q0 (singleton f0) delta w) ->
  (wordset_star l1 w).

(* Then we prove that the recognized language is l1* . *)

Lemma star_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset (wordset_star l1)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).

End Aut_for_star.

(* We prove here that the family of regular sets is closed under
   the Kleene closure. *)

Theorem closed_under_star :
  (l:(wordset A))
  (regexp I l)
    ->(nfe'_regularset nat eps I l)
      ->(nfe'_regularset nat eps I (wordset_star l)).

End Star.
