(**************************************************************************)
(*            The family of regular sets if closed under union            *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.
Require DisjUnion.
Require NatProperties.


(* We prove here that the family of regular sets is closed under union 	*)

Section Union.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Hypothesis eps_not_in_I : ~(member eps I).


(* Here we define the NFE' automaton that recognizes
   the union of l1 and l2. *)

Section Aut_for_union.

Variable l1,l2 : (wordset A).

Variable Q1,Q2 : (fset nat).
Variable q0_1,q0_2 : nat.
Variable f1,f2 : nat.
Variable delta1,delta2 : (fset nat*A*nat).

Hypothesis aut1 : (nfe'_automaton eps I Q1 q0_1 f1 delta1).
Hypothesis aut2 : (nfe'_automaton eps I Q2 q0_2 f2 delta2).

(* First we construct the elements of the automaton
   (Q,q0,f0 and delta). *)

Definition Q1Q2 : (fset nat*bool) := (disj_union Q1 Q2).

Definition exists_q0 := (outside2 Q1Q2).

Definition q0 : nat*bool := (proj1_sig ? ? exists_q0).

Definition exists_f0 := (outside2 (add q0 Q1Q2)).

Definition f0 : nat*bool := (proj1_sig ? ? exists_f0).

Definition Q : (fset nat*bool) := (add q0 (add f0 Q1Q2)).

Definition delta'1 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (linj q,(a,linj q')) end delta1).

Definition delta'2 :=
  (map [x:nat*A*nat]Cases x of (q,(a,q')) => (rinj q,(a,rinj q')) end delta2).

Definition delta : (fset (nat*bool)*A*(nat*bool)) :=
  (add (q0,(eps,(linj q0_1))) 
  (add (q0,(eps,(rinj q0_2)))
  (add (linj f1,(eps,f0))
  (add (rinj f2,(eps,f0))
  (union delta'1 delta'2))))).


(* Then we prove that it is a NFE' automaton. *)

Lemma member_f0_Q : (member f0 Q).
Proof.
Unfold f0 Q. Auto.
Save.

Lemma member_q0_Q : (member q0 Q).
Proof.
Unfold q0 Q. Auto.
Save.

Lemma f0_not_q0 : ~f0=q0.
Proof.
Generalize (sig_sig exists_f0) .
Change ~(member f0 (add q0 Q1Q2))->~f0=q0.
Unfold not.
Intros H1 equ.
Apply H1.
Rewrite -> equ.
Auto.
Save.
Hint f0_not_q0.

Lemma f0_not_linj_q1 : (q:nat)(member q Q1) -> ~f0=(linj q).
Proof.
Intros q Hq.
Generalize (sig_sig exists_f0) .
Change ~(member f0 (add q0 Q1Q2))->~f0=(linj q).
Unfold not.
Intros H1 equ.
Apply H1.
Apply member_x_add_y.
Unfold Q1Q2.
Rewrite -> equ.
Apply disj_union_l.
Assumption.
Save.

Lemma f0_not_linj_f1 : ~f0=(linj f1).
Proof.
Apply f0_not_linj_q1.
Elim aut1.
Auto.
Save.
Hint f0_not_linj_f1.

Lemma f0_not_rinj_q2 : (q:nat)(member q Q2) -> ~f0=(rinj q).
Proof.
Intros q Hq.
Generalize (sig_sig exists_f0) .
Change ~(member f0 (add q0 Q1Q2))->~f0=(rinj q).
Unfold not.
Intros H1 equ.
Apply H1.
Apply member_x_add_y.
Unfold Q1Q2.
Rewrite -> equ.
Apply disj_union_r.
Assumption.
Save.

Lemma f0_not_rinj_f2 : ~f0=(rinj f2).
Proof.
Apply f0_not_rinj_q2.
Elim aut2.
Auto.
Save.
Hint f0_not_rinj_f2.

Lemma f0_notin_delta'1 : (q:nat*bool)(a:A)
  ~(member (f0,(a,q)) delta'1).
Proof.
Intros q a.
Unfold delta'1.
Red.
Intros Hm.
Elim (map_def_2 Hm).
Intros x Hc.
Elim Hc.
Clear  Hm Hc.
Case x.
Intros n p.
Case p.
Intros a0 n0 Hm.
Intros equ.
Injection equ.
Intros.
Absurd f0=(linj n).
Apply f0_not_linj_q1.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a0,n0).
Apply member_subset with a:=delta1.
Assumption.
 
(Elim aut1; Auto).
 
Assumption.
Save.

Lemma f0_notin_delta'2 : (q:nat*bool)(a:A)
  ~(member (f0,(a,q)) delta'2).
Proof.
Intros q a.
Unfold delta'2.
Red.
Intros Hm.
Elim (map_def_2 Hm).
Intros x Hc.
Elim Hc.
Clear  Hm Hc.
Case x.
Intros n p.
Case p.
Intros a0 n0 Hm.
Intros equ.
Injection equ.
Intros.
Absurd f0=(rinj n).
Apply f0_not_rinj_q2.
Apply prod_left with b:=(product (add eps I) Q2) xb:=(a0,n0).
Apply member_subset with a:=delta2.
Assumption.
 
(Elim aut2; Auto).
 
Assumption.
Save.

Lemma H_f0 : (a:A)(q:nat*bool)~(member (f0,(a,q)) delta).
Proof.
Intros a q.
Red.
Unfold delta.
Intros Hm. Elim (add_def_1 Hm).

Intros equ.
Injection equ.
Intros.
(Absurd f0=q0; Auto).

Clear Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Injection equ.
Intros.
(Absurd f0=q0; Auto).

Clear Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Injection equ.
Intros.
Absurd f0=(linj f1) ; Auto.

Clear Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Injection equ.
Intros.
Absurd f0=(rinj f2) ; Auto.

Clear Hm. Intro Hm. Elim (union_def_1 Hm).
Change ~(member (f0,(a,q)) delta'1).
Apply f0_notin_delta'1.
Change ~(member (f0,(a,q)) delta'2).
Apply f0_notin_delta'2.
Save.

Lemma H_delta : (subset delta (product Q (product (add eps I) Q))).
Proof.
Unfold subset delta.
Intros x Hm. Elim (add_def_1 Hm).

Clear  Hm.
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Exact member_q0_Q. 
Apply member_pair_prod.
Auto.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_l.
(Elim aut1; Auto).

Clear  Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Exact member_q0_Q. 
Apply member_pair_prod.
Auto.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_r.
(Elim aut2; Auto).

Clear  Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_l.
(Elim aut1; Auto).
Apply member_pair_prod.
Auto.
Exact member_f0_Q.

Clear  Hm. Intro Hm. Elim (add_def_1 Hm).
Intros equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_r.
(Elim aut2; Auto).
Apply member_pair_prod.
Auto.
Exact member_f0_Q.

Clear  Hm. Intro Hm. Elim (union_def_1 Hm).

Unfold delta'1.
Clear  Hm. Intros Hm.
Elim (map_def_2 Hm).
Clear  Hm.
Intros x0.
Case x0.
Intros n p.
Case p.
Intros a n0.
Intros Hc. Elim Hc. Clear  Hc. Intros Hm equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_l.
Apply prod_left with b:=(product (add eps I) Q1) xb:=(a,n0).
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).
Apply member_pair_prod.
Apply prod_left with b:=Q1 xb:=n0.
Apply prod_right with a:=Q1 xa:=n.
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_l.
Apply prod_right with a:=(add eps I) xa:=a.
Apply prod_right with a:=Q1 xa:=n.
Apply member_subset with a:=delta1.
Assumption.
(Elim aut1; Auto).

Unfold delta'2.
Clear  Hm. Intros Hm.
Elim (map_def_2 Hm).
Clear  Hm.
Intros x0.
Case x0.
Intros n p.
Case p.
Intros a n0.
Intros Hc. Elim Hc. Clear  Hc. Intros Hm equ.
Rewrite -> equ.
Apply member_pair_prod.
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_r.
Apply prod_left with b:=(product (add eps I) Q2) xb:=(a,n0).
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Apply member_pair_prod.
Apply prod_left with b:=Q2 xb:=n0.
Apply prod_right with a:=Q2 xa:=n.
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Unfold Q.
Apply member_x_add_y.
Apply member_x_add_y.
Unfold Q1Q2.
Apply disj_union_r.
Apply prod_right with a:=(add eps I) xa:=a.
Apply prod_right with a:=Q2 xa:=n.
Apply member_subset with a:=delta2.
Assumption.
(Elim aut2; Auto).
Save.

Lemma union_automaton : (nfe'_automaton eps I Q q0 f0 delta).
Proof.
Apply (nfe'_automaton_intro 
      	 member_f0_Q
      	 eps_not_in_I 
      	 member_q0_Q
      	 H_f0
      	 H_delta).
Save.

(* Path extension. 
   If we have a path in automaton A1 or A2, then we have the corresponding
   path in the automaton A. *)

Axiom path_extension_1 : 
  (w:(word A))(q,q':nat)
  (nfe_path eps I Q1 delta1 q        q'        w) ->
  (nfe_path eps I Q  delta  (linj q) (linj q') w).

Axiom path_final_extension_1 : 
  (w:(word A))
  (nfe_path eps I Q delta (linj q0_1) (linj f1) w) ->
  (nfe_path eps I Q delta q0          f0        w).
  
Axiom path_extension_2 : 
  (w:(word A))(q,q':nat)
  (nfe_path eps I Q2 delta2 q        q'        w) ->
  (nfe_path eps I Q  delta  (rinj q) (rinj q') w).

Axiom path_final_extension_2 : 
  (w:(word A))
  (nfe_path eps I Q delta (rinj q0_2) (rinj f2) w) ->
  (nfe_path eps I Q delta q0          f0        w).


(* Path restriction.
   If we have a path in the new automaton from q0 to f0 for w,
   then there is either a path from q0_1 to f1 in A1 for w
   or a path from q0_2 to f2 in A2. *)

Axiom path_restriction : 
  (w:(word A))
  (nfe_path eps I Q delta q0 f0 w) ->
  (nfe_path eps I Q1 delta1 q0_1 f1 w) \/ (nfe_path eps I Q2 delta2 q0_2 f2 w).

(* Then we prove that the recognized language is the union of l1 and l2. *) 

Lemma union_is_reg : 
  (eq_wordset l1 (nfe_accepts eps I Q1 q0_1 (singleton f1) delta1)) ->
  (eq_wordset l2 (nfe_accepts eps I Q2 q0_2 (singleton f2) delta2)) ->
  (eq_wordset (wordset_union l1 l2)
              (nfe_accepts eps I Q q0 (singleton f0) delta)).
Proof.
Intros H_l1 H_l2.
Unfold eq_wordset.
Intros w.
Split.

(* 1. if a word w is in the union of l1 and l2 then it is accepted. *)
Intros Hu. Elim Hu.

Clear  Hu w. Intros w l1_w.
Unfold eq_wordset in H_l1.
Elim (H_l1 w).
Clear  H_l1.
Intros H_l1_1 H_l1_2.
Generalize (H_l1_1 l1_w) .
Intros acc1. Elim acc1.
Intros f H_f Hpath.
Apply nfe_accepts_intro with f:=f0.
Auto.
Generalize (member_of_singleton H_f) .
Intros equ.
Rewrite equ in Hpath.
Generalize (path_extension_1 Hpath) .
Intros Hpath'.
Apply path_final_extension_1.
Assumption.
 
Clear  Hu w. Intros w l2_w.
Unfold eq_wordset in H_l2.
Elim (H_l2 w).
Clear  H_l2.
Intros H_l2_1 H_l2_2.
Generalize (H_l2_1 l2_w) .
Intros acc2. Elim acc2.
Intros f H_f Hpath.
Apply nfe_accepts_intro with f:=f0.
Auto.
Generalize (member_of_singleton H_f) .
Intros equ.
Rewrite equ in Hpath.
Generalize (path_extension_2 Hpath) .
Intros Hpath'.
Apply path_final_extension_2.
Assumption.
 
(* 2. if a word is accepted then it is in the union of l1 and l2. *)
Intros Hacc.
Elim Hacc.
Intros f Hf.
Generalize (member_of_singleton Hf) .
Intros equ.
Rewrite -> equ.
Intros Hpath.
Generalize (path_restriction Hpath) .
Intros disj. Elim disj.

Intros path1.
Unfold eq_wordset in H_l1.
Elim (H_l1 w).
Intros Hl11 Hl12.
Apply wordset_union_g.
Apply Hl12.
Apply nfe_accepts_intro with f:=f1.
Auto.
Assumption.

Intros path2.
Unfold eq_wordset in H_l2.
Elim (H_l2 w).
Intros Hl21 Hl22.
Apply wordset_union_d.
Apply Hl22.
Apply nfe_accepts_intro with f:=f2.
Auto.
Assumption.
Save.


End Aut_for_union.


Theorem closed_under_union :
  (l1,l2:(wordset A))
  (regexp I l1)
   ->(nfe'_regularset nat eps I l1)
      ->(regexp I l2)
         ->(nfe'_regularset nat eps I l2)
      	   ->(nfe'_regularset nat*bool eps I (wordset_union l1 l2)).
Proof.
Intros l1 l2 reg_l1 nfe'_l1 reg_l2 nfe'_l2.
Elim nfe'_l1. Intros Q1 q0_1 f1 delta1 aut1 H_l1.
Elim nfe'_l2. Intros Q2 q0_2 f2 delta2 aut2 H_l2.
Apply nfe'_regular_intro
  with Q  := (Q Q1 Q2)
       q0 := (q0 Q1 Q2)
       f  := (f0 Q1 Q2)
       delta := (delta Q1 Q2 q0_1 q0_2 f1 f2 delta1 delta2).
Apply union_automaton ; Assumption.
Apply union_is_reg ; Assumption.
Save.

End Union.
