(**************************************************************************)
(*                     The empty wordset is regular.                      *)
(**************************************************************************)

Implicit Arguments On.

Require NFA_eps.

Section Empty_language.

Variable A : Set.
Variable I : (fset A).
Variable eps : A.
Variable H_eps : ~(member eps I).

(* First, we define the automaton (NFE') that recognizes the empty language.
   It is an automaton with two states, 0 and 1, 0 being the initial state
   and f the final state. *)

Local Q  := (add (S O) (singleton O)).
Local q0 := O.
Local f  := (S O).
Local delta := (empty_set nat*A*nat).

(* Then we prove that it is a NFE' automaton. (The proofs are all 
   trivial with this very simple automaton). *)

Local member_f_Q : (member f Q).
Proof. Unfold Q. Auto. Save.

Local member_q0_Q : (member q0 Q).
Proof. Unfold q0. Unfold Q. Auto. Save.

Local H_f : (a:A)(q:nat) ~(member (f,(a,q)) delta).
Proof. Unfold delta. Auto. Save.

Local H_delta : (subset delta (product Q (product (add eps I) Q))).
Proof. Unfold delta. Auto. Save.

Local is_automaton : (nfe'_automaton eps I Q q0 f delta).
Proof.
Apply (nfe'_automaton_intro
        member_f_Q
	H_eps
	member_q0_Q
	H_f H_delta).
Save.

(* Then we prove that this automaton recognizes exactly the empty
   language. *)

Local eq_wordset : (eq_wordset [w:(word A)]False
      	       	       	       (nfe_accepts eps I Q q0 (singleton f) delta)).
Proof.
Unfold eq_wordset.
Intros w.
Split.
Intro.
Elim H.

Intros H_accepts.
Elim H_accepts.
Intros f' H_f' H_path.
Inversion H_path.
Unfold q0 in H1.
Unfold f in H_f'.
Rewrite <- H1 in H_f'.
Generalize (add_def_1 H_f') .
Intros disj.
Elim disj.
Intros O_SO.
(Absurd O=(S O); Auto).

Change ~(member O (empty_set nat)).
Auto.

Unfold delta.
(Absurd (member (q0,(a,e1)) delta); Auto).
Unfold delta.
Auto.

Unfold delta in H0.
(Absurd (member (q0,(eps,e1)) (empty_set nat*A*nat)); Auto).
Save.

(* And finally, as a consequence of the two previous results, we
   prove that the empty language is regular. *)

Lemma empty_regular : (nfe'_regularset nat eps I [w:(word A)]False ).

End Empty_language.
