(**************************************************************************)
(*                            Kleene's theorem                            *)
(**************************************************************************)

Implicit Arguments On.

Require AutEquiv.
Require NatBoolNat.

Require Empty.
Require OneWord.
Require Closed_union.
Require Closed_concat.
Require Closed_star.

(* Let A be the type of symbols, I a finite set of symbols (the alphabet)
   and eps an element of type A which does not belong to I. *)

Parameter A : Set.
Parameter I : (fset A).
Parameter eps : A.
Axiom eps_not_in_I : ~(member eps I).

(* The family of regular sets (over A) contains the empty language,
   all {w} for w in A* and is closed under union, concatenation and
   Kleene closure. *)

(* First we construct a NFE' automaton for any regular expression. *)

Theorem regexp_nfe'_automaton : (L:(wordset A))
    (regexp I L) -> (nfe'_regularset nat eps I L).

(* Then we construct a DFA automaton using the equivalence between
   NFE' and DFA. *)

Theorem regexp_automaton : (L:(wordset A))
    (regexp I L) -> (regularset nat I L).

