(**************************************************************************)
(*                            Kleene's theorem                            *)
(**************************************************************************)

Implicit Arguments On.

Require AutEquiv.
Require NatBoolNat.

Require Empty.
Require OneWord.
Require Closed_union.
Require Closed_concat.
Require Closed_star.


(* Let A be the type of symbols, I a finite set of symbols (the alphabet)
   and eps an element of type A which does not belong to I. *)

Parameter A : Set.
Parameter I : (fset A).
Parameter eps : A.
Axiom eps_not_in_I : ~(member eps I).


(* The family of regular sets (over A) contains the empty language,
   all {w} for w in A* and is closed under union, concatenation and
   Kleene closure. *)

(* First we construct a NFE' automaton for any regular expression. *)

Theorem regexp_nfe'_automaton : (L:(wordset A))
    (regexp I L) -> (nfe'_regularset nat eps I L).

Proof.
Intros L L_reg.
Elim L_reg.

(* 1. empty language *)
  Apply (empty_regular eps_not_in_I).

(* 2. language reduced to one word *)
  Intros w H_w.
  Apply (one_word_regular eps_not_in_I H_w).

(* 3. union *)
  Intros l1 l2 reg_l1 nfe'_l1 reg_l2 nfe'_l2.
  Generalize (closed_under_union eps_not_in_I reg_l1 nfe'_l1 reg_l2 nfe'_l2).
  Intro H.
  Apply (natbool_equiv_nat eps_not_in_I H).

(* 4. concatenation *)
  Intros l1 l2 reg_l1 nfe'_l1 reg_l2 nfe'_l2.
  Generalize (closed_under_concat eps_not_in_I reg_l1 nfe'_l1 reg_l2 nfe'_l2).
  Intro H.
  Apply (natbool_equiv_nat eps_not_in_I H).

(* 5. Kleene closure *)
  Intros l reg_l nfe'_l.
  Apply (closed_under_star eps_not_in_I reg_l nfe'_l).

Save.


(* Then we construct a DFA automaton using the equivalence between
   NFE' and DFA. *)

Theorem regexp_automaton : (L:(wordset A))
    (regexp I L) -> (regularset nat I L).

Proof.
Intros L L_reg.
Apply NFE'_equiv_DFA with eps:=eps.
Exact eps_not_in_I.
Exact (regexp_nfe'_automaton L_reg).
Save.

