(**************************************************************************)
(*                             Disjoint union                             *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.
Require Properties.

(* Definition of disjoint union : 
   We define the disjoint union of a and b as 

      { (x,false) | x in a}  U  { (x,true) | x in b }

   Therefore, the disjoint union of two sets over type A is a set
   over type A*bool. *)

Definition linj : (A:Set)A->A*bool := [A:Set][x:A](x,false).

Definition rinj : (A:Set)A->A*bool := [A:Set][x:A](x,true).

Definition disj_union : (A:Set)(fset A) -> (fset A) -> (fset A*bool) :=
  [A:Set][a,b:(fset A)] (union (map [x:A](linj x) a) (map [x:A](rinj x) b)).

(* Properties of disjoint union : *)

Section properties.

Variable A : Set.
Variable a,b,c,d : (fset A).

Lemma member_map_linj : (x:A*bool)
  (member x (map [x:A](linj x) a)) -> (member (Fst x) a).

Lemma member_map_rinj : (x:A*bool)
  (member x (map [x:A](rinj x) a)) -> (member (Fst x) a).

Lemma absurd_linj_rinj : (x:A*bool)
  (member x (map [x:A](linj x) a)) ->
  (member x (map [x:A](rinj x) b)) ->
  False.

Lemma member_disj_union : (y:A*bool)
  (member y (disj_union a b)) ->
     (Ex [x:A]((member x a) /\ y=(linj x))) 
  \/ (Ex [x:A]((member x b) /\ y=(rinj x))).

Lemma disj_union_l : (x:A)
  (member x a) -> (member (linj x) (disj_union a b)).

Lemma disj_union_r : (x:A)
  (member x b) -> (member (rinj x) (disj_union a b)).

End properties.

