(**************************************************************************)
(*                             Disjoint union                             *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.
Require Properties.


(* Definition of disjoint union : 
   We define the disjoint union of a and b as 

      { (x,false) | x in a}  U  { (x,true) | x in b }

   Therefore, the disjoint union of two sets over type A is a set
   over type A*bool. *)

Definition linj : (A:Set)A->A*bool := [A:Set][x:A](x,false).

Definition rinj : (A:Set)A->A*bool := [A:Set][x:A](x,true).

Definition disj_union : (A:Set)(fset A) -> (fset A) -> (fset A*bool) := 
  [A:Set][a,b:(fset A)] (union (map [x:A](linj x) a) (map [x:A](rinj x) b)).



(* Properties of disjoint union : *)

Section properties.

Variable A : Set.
Variable a,b,c,d : (fset A).


Lemma member_map_linj : (x:A*bool)
  (member x (map [x:A](linj x) a)) -> (member (Fst x) a). 
Proof.
Intros x H.
Elim (map_def_2 H).
Intros x0 H0.
Elim H0.
Intros Hm equ.
Rewrite -> equ.
Auto.
Save.


Lemma member_map_rinj : (x:A*bool)
  (member x (map [x:A](rinj x) a)) -> (member (Fst x) a). 
Proof.
Intros x H.
Elim (map_def_2 H).
Intros x0 H0.
Elim H0.
Intros Hm equ.
Rewrite -> equ.
Auto.
Save.


Lemma absurd_linj_rinj : (x:A*bool)
  (member x (map [x:A](linj x) a)) ->
  (member x (map [x:A](rinj x) b)) ->
  False.
Proof.
Intros x Hl Hr.
Elim (map_def_2 Hl).
Intros x0 H0.
Elim H0.
Intros H1 H2.
Elim (map_def_2 Hr).
Intros x1 H3.
Elim H3.
Intros H4 H5.
(Absurd true=false; Auto).
Discriminate.
 
Replace true with (Snd (rinj x1)).
Replace false with (Snd (linj x0)).
Rewrite <- H2.
Rewrite <- H5.
Reflexivity.
 
Auto.
 
Auto.
Save.


Lemma member_disj_union : (y:A*bool)
  (member y (disj_union a b)) ->
     (Ex [x:A]((member x a) /\ y=(linj x))) 
  \/ (Ex [x:A]((member x b) /\ y=(rinj x))).
Proof.
Intros y Hy.
Elim (union_def_1 Hy).
Intros Hl.
Left.
Elim (map_def_2 Hl).
Auto.
Intros x Hx.
Exists  x.
Auto.
 
Intros Hr.
Right.
Elim (map_def_2 Hr).
Intros x Hx.
Exists  x.
Auto.
Save.


Lemma disj_union_l : (x:A)
  (member x a) -> (member (linj x) (disj_union a b)).
Proof.
Intros x Hx.
Unfold disj_union.
Apply union_def_2.
Left.
Apply map_def_3.
Exists  x.
Auto.
Save.


Lemma disj_union_r : (x:A)
  (member x b) -> (member (rinj x) (disj_union a b)).
Proof.
Intros x Hx.
Unfold disj_union.
Apply union_def_2.
Right.
Apply map_def_3.
Exists  x.
Auto.
Save.


End properties.

