
(**************************************************************************)
(*                  Definition of maximum and properties                  *)
(**************************************************************************)

Require Le.
Require Lt.


Fixpoint max [n:nat] : nat -> nat :=
  [m:nat] 
        (Cases n m of
	    O  _ => m
	 | (S n') O => (S n')
	 | (S n') (S m') => (S (max n' m')) 
         end).

Lemma max_SS : (n,m:nat)(S (max n m))=(max (S n) (S m)).
Auto.
Save.

Lemma le_max_l : (n,m:nat)(le n (max n m) ).
Induction n; Intros; Simpl; Auto.
Elim m; Intros; Simpl; Auto.
Save.
Hint le_max_l.

Lemma le_max_r : (n,m:nat)(le m (max n m) ).
Induction n; Simpl; Auto.
Induction m; Simpl; Auto.
Save.
Hint le_max_r.

Lemma lt_max_l : (x,y,z:nat) (lt x y) -> (lt x (max y z)).
Intros.
Apply lt_le_trans with y; Auto.
Save.
Hint lt_max_l.

Lemma lt_max_r : (x,y,z:nat) (lt x z) -> (lt x (max y z)).
Intros.
Apply lt_le_trans with z; Auto.
Save.
Hint lt_max_r.

Lemma le_trans_max_r : (x,y,z:nat) (le x z) -> (le x (max y z)).
Intros.
Apply le_trans with z; Auto.
Save.
Hint le_trans_max_r.


(* If a property holds for both n and m then it holds for (max n m). *)

Lemma max_case : (n,m:nat)(P:nat->Prop)(P n)->(P m)->(P (max n m)).
Proof.
Induction n; Simpl; Auto.
Induction m; Intros; Simpl; Auto.
Pattern (max n0 n1); Apply H ; Auto.
Save.


Require PolyList.

(* The maximum element of a list *)

Fixpoint max_of_list [l:(list nat)] : nat :=
          (Cases l of
	     nil => O	
	  | (cons x l') => (max x (max_of_list l'))
          end ).

Lemma max_cons : (x:nat)(l:(list nat))
  (max_of_list (cons nat x l))=(max x (max_of_list l)).
Proof.
Intros x.
Induction l; Auto.
Save.
Hint max_cons.


(* Any element of a list l is lower or equal to (max_of_list l) *)

Lemma le_max_list : (l:(list nat))(n:nat) 
  (In nat n l) -> (le n (max_of_list l)).
Proof.
Induction l.

Intros.
Absurd (In nat n (nil nat)); Auto.
Red. Auto.

Intros a l' H n H0.
Simpl.
Elim H0.
Intros equ.
Rewrite <- equ.
Auto.

Intros H1.
Generalize (H n H1) .
Auto.
Save.


Lemma Smax_outside : (l:(list nat))
  ~(In nat (S (max_of_list l)) l).
Proof.
Intros l.
Red.
Intro H. 
Generalize (le_max_list l (S (max_of_list l)) H).
Change ~(le (S (max_of_list l)) (max_of_list l)).
Auto.
Save.

