(**************************************************************************)
(*                     Various lemmas on finite sets                      *)
(**************************************************************************)

Implicit Arguments On.

Require FiniteSets.

Section Properties_of_add.

Variable A : Set.
Variable a : (fset A).

Lemma add_def_1 : (x,y:A) (member x (add y a)) -> (x=y \/ (member x a)).

Lemma add_def_2 : (x,y:A) (x=y \/ (member x a)) -> (member x (add y a)).

Lemma member_x_add_x : (x:A) (member x (add x a)).

Lemma member_x_add_y : (x,y:A) (member x a) -> (member x (add y a)).

Lemma member_add_inv_1 : (x,y:A)
  (member x (add y a)) -> ~x=y -> (member x a).

End Properties_of_add.

(* a member to the singleton {a} *)

Lemma member_x_singleton_x : (A:Set) (a:A) (member a (singleton a)).

Lemma member_of_singleton : (A:Set) (a,x:A)
  (member x (singleton a)) -> x=a.

Section Properties_of_product.

Variable A,B : Set.
Variable a   : (fset A).
Variable b   : (fset B).

Lemma prod_def_1 : (x:A*B)
    (Cases x of (pair xa xb) => (member xa a)/\(member xb b) end)
 -> (member x (product a b)).

Lemma prod_def_2 : (x:A*B)
    (member x (product a b))
 -> (Cases x of (pair xa xb) => (member xa a)/\(member xb b) end).

Lemma member_pair_prod : (xa:A)(xb:B)
  (member xa a) -> (member xb b) ->
  (member (xa,xb) (product a b)).

Lemma prod_left : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) -> (member xa a).

Lemma prod_right : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) -> (member xb b).

Lemma member_pair_prod_inv : (xa:A)(xb:B)
  (member (xa,xb) (product a b)) ->
  (member xa a) /\ (member xb b).

End Properties_of_product.

Section Properties_of_subset.

Variable A,B : Set.
Variable a   : (fset A).
Variable b   : (fset B).

Lemma subset_a_a : (subset a a).

Lemma subset_a_addxa : (x:A) (subset a (add x a)).

Lemma subset_trans_add : (a':(fset A)) (x:A)
  (subset a a') -> (subset a (add x a')).

Lemma subset_add_trans : (a':(fset A)) (x:A)
  (member x a') -> (subset a a') -> (subset (add x a) a').

Lemma subset_empty_set : (subset (empty_set A) a).

Lemma subset_trans_product_left : (d:(fset A*B)) (x:A)
  (subset d (product a b)) ->
  (subset d (product (add x a) b)).

Lemma subset_trans_product_right : (d:(fset A*B)) (x:B)
  (subset d (product a b)) ->
  (subset d (product a (add x b))).

Lemma member_subset : (a':(fset A)) (x:A)
  (member x a) -> (subset a a') -> (member x a').

End Properties_of_subset.

Section Properties_of_eq_fset.

Variable A' : Set.
Variable a : (fset A').

Lemma eq_fset_refl : (eq_fset a a).

End Properties_of_eq_fset.

Lemma subset_singleton2 :
  (A:Set)(f:A)(F,Q:(fset A))
  (eq_fset F (singleton f)) -> 
  (member f Q) ->
  (subset F Q).

Lemma subset_singleton :
  (A:Set)(f:A)(Q:(fset A))
  (member f Q) ->
  (subset (singleton f) Q).

(* Properties of union *)

Lemma union_def_1 : (A:Set) (a,a':(fset A)) (x:A)
     (member x (union a a'))
  -> ((member x a) \/ (member x a')).

Lemma union_def_2 : (A:Set) (a,a':(fset A)) (x:A)
    ((member x a) \/ (member x a'))
 -> (member x (union a a')).

(* Properties of map *)

Lemma map_def_1 : (A,B:Set)(f:A->B)(a:(fset A))
  (x:A) (member x a) -> (member (f x) (map f a)).

Lemma map_def_2 : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (member y (map f a)) -> (Ex [x:A]((member x a) /\ y=(f x))).

Lemma map_def_3 : (A,B:Set)(f:A->B)(a:(fset A))
  (y:B) (Ex [x:A]((member x a) /\ y=(f x))) -> (member y (map f a)).

(* Properties of list_of_fset *)

Lemma list_of_def_1 : (A:Set)(a:(fset A))
  (x:A) (member x a) -> (In A x (list_of_fset a)).

Lemma list_of_def_2 : (A:Set)(a:(fset A))
  (x:A) (In A x (list_of_fset a)) -> (member x a).
